// C Source File
// Created 06-07-2002; 13:46:10

#include "konv.h"
#include "bg.h"

enum {X = 53, Y_UINT = 27, Y_INT = 41, Y_FRAC = 55, Y_BIN = 69, Y_HEX = 83};

// Sprites that make up the marker
static unsigned long marker1[]={0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF};
static unsigned int marker2[]={0xFFF0,0xFFF0,0xFFF0,0xFFF0,0xFFF0,0xFFF0,0xFFF0,0xFFF0,0xFFF0};

// Sprite for the blinking cursor
static unsigned char cur[]={0x60,0x60,0x60,0x60,0x60,0x60,0x60};



typedef struct {
	int x;
	int y;
	int on;
} CURSOR;

CURSOR cursor;


typedef struct{
	int x;
	int y;
}MARKER;

MARKER marker;

// Function to draw the marker consisting of two sprites
void DrawMarker(MARKER *m){
	Sprite32(m->x,m->y,9,marker1,LCD_MEM,SPRT_XOR);
	Sprite16(m->x+32,m->y,9,marker2,LCD_MEM,SPRT_XOR);
}
// Pointer to the DrawMarker function for easier reading of the code
void (*EraseMarker)(MARKER *m) = &DrawMarker;


// Five functions to tell if the key pressed is allowed
int TestKeyUnsignedInt(short key, unsigned int i){
	if(((key >= '0') && (key <= '9')) && i < 5)
		return 1;
	return 0;
}

int TestKeyInt(short key, unsigned int i){
	if(((key == KEY_SIGN && !i) || ((key >= '0') && (key <= '9'))) && i < 6)
		return 1;
	return 0;
}

int TestKeyFrac(short key, unsigned int i){
	if((((key >= '0') && (key <= '9')) || (key == KEY_SIGN) || (key == '.') || (key == 0x95)) && i < 16)
		return 1;
	return 0;
}

int TestKeyBin(short key, unsigned int i){
	if(((key == '0') || (key == '1')) && i < 16)
		return 1;
	return 0;
}

int TestKeyHex(short key, unsigned int i){
	if((((key >= '0') && (key <= '9')) || ((key >= 'a') && (key <= 'f'))) && i < 4)
		return 1;
	return 0;
}
// An array of pointers to the above five test functions
int (*TestKey[])(short key, unsigned int i) = {&TestKeyUnsignedInt,&TestKeyInt,&TestKeyFrac,&TestKeyBin,&TestKeyHex};

// Displays the error message " OUT OF RANGE "
void Err_Out_of_Range(short x, short y, char *str){
	DrawStr(x,y,"Out of range",A_XOR);
	ngetchx();
	DrawStr(x,y,"Out of range",A_XOR);		
	DrawStr(x,y,str,A_REPLACE);					
}

// Five functions to process the string entered. The sting is processed differently
// depending on the type: Unsigned Int, Signed Int, Fractional, Binary, Hexadecimal.
int ProcessStrUnsignedInt(int *number, char *str){
	long num;
	
		if((num = strtol(str,NULL,0)) > 65535){
			Err_Out_of_Range(X,Y_UINT,str);
		}else{
			*number = num;
			return 1;
		}
	return 0;
}

int ProcessStrInt(int *number,char *str){
	long num;

		num = strtol(str,NULL,0);
		if((num < -32768) || (num > 32767)){
			Err_Out_of_Range(X,Y_INT,str);
		}else{
			*number = num;
			return 1;
		}
	return 0;
}

int ProcessStrFrac(int *number, char *str){
	float num;
	
	if(is_nan(num = atof(str))){
			DrawStr(X,Y_FRAC,"Styntax",A_XOR);
			ngetchx();
			DrawStr(X,Y_FRAC,"Styntax",A_XOR);		
			DrawStr(X,Y_FRAC,str,A_REPLACE);		
		}else{ 
			num = num * 32768;
			if((num < -32768) || (num > 32767.0)){
				Err_Out_of_Range(X,Y_FRAC,str);
			}else{
				push_shortint(0);
				push_Float(num);
				push_quantum(ROUND_TAG);
				NG_approxESI(top_estack);
				estack_to_short(top_estack,number);
				top_estack -= 23;				
				return 1;
			}
		}
				
	return 0;
}

int ProcessStrBin(int *number, char *str){
	unsigned int i;
	unsigned int num = 0;
		
	for(i=0;i<strlen(str);i++){
		num <<= 1;
	if(str[i] == '1')
		num++;
	}
	*number = num;
	return 1;
}

int ProcessStrHex(int *number, char *str){
	char tmp[7] = "0x", c;
	int i;

	for(i = 0; (c = str[i]); i++){
		tmp[i+2] = c;
	}
	tmp[i+2] = 0;
	
	*number = (int)strtol(tmp,NULL,0);
	return 1;
}
// An array of pointers to the above five functions
int (*ProcessStr[])(int *number, char *str) = {&ProcessStrUnsignedInt,&ProcessStrInt,&ProcessStrFrac,&ProcessStrBin,&ProcessStrHex};

// The GetNumber function gets a string while displaying a blinking cursor at the entrypoint.
// It decides whitch type of number it is getting using the nr argument and decides whether 
// the key presses are allowed by calling the appropriate TestKey-function. When the string
// is terminated by pressing ENTER, the sting is converted to an int. This is done by calling 
// the appropriate ProcessStr-function again using the nr argument to decide whitch.
int GetNumber(int nr, CURSOR *p, short k){
	short	 key, x = p->x, y = p->y;
	char str[18] = "\0", done = 0;
	unsigned int i = 0;
	int ok = 0, number = 0;
	void *kbq = kbd_queue();

	if(k != KEY_ENTER){
		str[i] = k; str[i+1] = 0;
		i++;
		p->x += 6;
		DrawStr(x,y,str,A_REPLACE);
	}
	
	Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
	p->on ^= 1;
	OSFreeTimer(USER_TIMER);
	OSRegisterTimer(USER_TIMER,12);
	while(!ok){
		done = 0;
		while(!done){
			if (!OSdequeue(&key,kbq)){
				key &= 0xF7FF;
				
				if(TestKey[nr](key,i)){
					str[i] = key; str[i+1] = 0;
					i++;
					if(p->on){
						Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
						p->x += 6;
						Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
					}else{
						p->x += 6;
					}
				}
				
				if((key == KEY_BACKSPACE) && i){
					DrawStr(x,y,str,A_XOR);
					i--;
					str[i] = 0;
					if(p->on){
						Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
						p->x -= 6;
						Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
					}else{
						p->x -= 6;
					}
				}
				
				if(key == KEY_ENTER){
					done = 1;
					if(!i){
						str[0] = '0';
						str[1] = 0;
					}
				}
				
				if(key == KEY_ESC){
					DrawStr(x,y,str,A_XOR);
					if(p->on){
						Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
						p->on ^= 1;
					}
					p->x = x;				
					return 0;
				}
				
				DrawStr(x,y,str,A_REPLACE);
			}
			
			if(OSTimerExpired(USER_TIMER)){
				Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
				p->on ^= 1;
			}
			
		}
		
		DrawStr(x,y,str,A_XOR);
		if(p->on){
			Sprite8(p->x,p->y,7,cur,LCD_MEM,SPRT_XOR);
			p->on ^= 1;				
		}
	
		if(ProcessStr[nr](&number,str)){
			ok = 1;
			p->x = x;
		}
	}

	return(number);	
}

// helpful function to display an int in binary format
void binprint(int a){
	int i;
	int n = sizeof(int) * CHAR_BIT;
	int mask = 1 << (n-1);
	
	for (i = 1; i <= n; i++){
		putchar(((a & mask) == 0) ? '0' : '1');
		a <<= 1;
		if ((i % CHAR_BIT == 0) && (i < n))
			putchar(' ');
	}
	putchar('\n');
}

// displays an int in hexadecimal format. printf didn't do the job I wanted.
// I wanted hex number to always be printed with 4 digits ( zero's in the most
// significant digits when the numbers are small).
void hexprint(unsigned int b){
	int i;
	int n = sizeof(int) * CHAR_BIT / 4;
	int mask = 0xF << ((n-1) * 4);

	for(i = 1; i <= n; i++){
		printf("%X",((b & mask) >> 12));
		b <<= 4;
	}	
}

// Displays num in the five different formats: Unsigned Int, Signed Int, Fractional, binary, Hexadecimal.
void printnumbers(int num){
	MoveTo(X,Y_UINT);
	printf("%u",num);
	MoveTo(X,Y_INT);
	printf("%d",num);
	MoveTo(X,Y_FRAC);
	printf("%y",fdiv(num,32768));
	MoveTo(X,Y_BIN);
	binprint(num);
	MoveTo(X,Y_HEX);
	hexprint(num);
}

// Main Function
void _main(void){
	int number = 0, done = 0, nr = 0;
	short key = 0;
	void *kbq = kbd_queue();

// Background text and frame
	clrscr();
	DrawBackground(bg,LCD_MEM);

	FontSetSys(F_4x6);
	DrawStr(5,Y_UINT+2,"Unsigned int",A_NORMAL);
	DrawStr(5,Y_INT+2,"Signed int",A_NORMAL);
	DrawStr(5,Y_FRAC+2,"Fractional",A_NORMAL);
	DrawStr(5,Y_BIN+2,"Binary",A_NORMAL);
	DrawStr(5,Y_HEX+2,"Hexadecimal",A_NORMAL);
	FontSetSys(F_6x8);

// Initialise the cursor
	cursor = (CURSOR){X,Y_UINT,0};

// Initialise and draw the marker
	marker = (MARKER){5,Y_UINT};
	DrawMarker(&marker);
	
// The main loop
	while(!done){

		if (!OSdequeue(&key,kbq))
			key &= 0xF7FF;
		else
			key = 0;

	
		if(key == KEY_ESC){
			done = 1;
		}else	if((key == KEY_UP) && (marker.y != Y_UINT)){
			EraseMarker(&marker);
			marker.y -= 14;
			DrawMarker(&marker);
			cursor.y -= 14;
			nr--;
			
		}else	if((key == KEY_DOWN) && (marker.y != Y_HEX)){
			EraseMarker(&marker);
			marker.y += 14;
			DrawMarker(&marker);
			cursor.y += 14;
			nr++;
		}else	if((key == KEY_ENTER) || TestKey[nr](key,0)){ 
			ScrRectFill(&(SCR_RECT){{X,Y_UINT,155,Y_HEX+6}},ScrRect,A_REVERSE);
			EraseMarker(&marker);
			number = GetNumber(nr,&cursor,key);
			printnumbers(number);
			DrawMarker(&marker);
		}

	}

}